// (c) Copyright 2012 Cloudera, Inc. All rights reserved.
define([
  "underscore",
  "cloudera/Util",
  "cloudera/cmf/charts/TimeSeriesVisualizer"
], function (_, Util, TimeSeriesVisualizer)  {

  /**
   * Renders a specific plot.
   * options {
   *   plot:           (required) the plot object.
   *   container:      (required) "selector or element of the container DOM object",
   *   context:        (optional) a mapping of placeholder variable names to values.
   *   resizeContainer:(optional) an element that can resize charts.
   *   resizeLabel:    (optional) an element that shows the width after resize.
   *   enableEditing:  (optional) true|false, whether editing should be enabled, default false.
   *   enableCloning:  (optional) true|false, whether cloning should be enabled, default false.
   *   enableRemoving: (optional) true|false, whether removing should be enabled, default false.
   *   enableFacets:   (optional) true|false, whether facets should be enabled, default false.
   *   enableEnlarging:(optional) true|false, whether the chart can be enlarged, default true.
   * }
   */
  var PlotContainer = function(options) {

    var self = this, $container = $(options.container);

    // Used for creating visualizers below.
    var args = {
      plot: options.plot,
      container: $container,
      context: options.context,
      resizeContainer: options.resizeContainer,
      resizeLabel: options.resizeLabel,
      enableEditing: options.enableEditing,
      enableCloning: options.enableCloning,
      enableRemoving: options.enableRemoving,
      enableFacets: options.enableFacets,
      enableEnlarging: options.enableEnlarging
    };

    var shouldShowPlot = function(data) {
      if (options.plot.hideIfNoSeries) {
        if (data.timeSeries.length === 0) {
          return false;
        }
        // BulkTsqueryFetcher pads empty responses with fake data, but marks
        // the data as fake. If we find fake data and the plot would like us
        // to hide it, then do so.
        if (data.timeSeries.length === 1 &&
            data.timeSeries[0].data.length === 2 &&
            data.timeSeries[0].data[0].type === 'fake' &&
            data.timeSeries[0].data[1].type === 'fake') {
          return false;
        }
      }
      return true;
    };

    self.initialize = function() {
      self.visualizer = self.createVisualizer(options.plot.chartType);
      if (_.isNumber(options.plot.width) && _.isNumber(options.plot.height)) {
        self.setDimension(options.plot.width, options.plot.height);
      }
      self.setYRange(options.plot.ymin, options.plot.ymax);
    };

    self.getViewModel = function() {
      return self.visualizer.getViewModel();
    };

    /**
     * Renders data.
     */
    self.render = function(data) {
      if (!shouldShowPlot(data)) {
        $container.hide();
        return;
      }
      if (data) {
        self.visualizer.render(data);
      }
    };

    /**
     * @return the state of the plot.
     */
    self.getPlot = function() {
      return self.getViewModel().getPlot();
    };

    /**
     * Sets the dimension of each chart.
     */
    self.setDimension = function(width, height) {
      self.getViewModel().setDimension(width, height);
    };

    /**
     * Sets the chart type.
     */
    self.setChartType = function(chartType) {
      if (!self.getViewModel().setChartType(chartType)) {
        // we need to switch visualizer.
        if (self.visualizer) {
          self.visualizer.unsubscribe();
        }
        self.visualizer = self.createVisualizer(chartType);
      }
    };

    self.setYRange = function(ymin, ymax) {
      self.getViewModel().setYRange(ymin, ymax);
    };

    self.createVisualizer = function(chartType) {
      return new TimeSeriesVisualizer(args);
    };

    /**
     * Sets the tsquery for the entire plot.
     */
    self.setTsquery = function(tsquery) {
      self.getViewModel().tsquery(tsquery);
    };

    /**
     * Gets the current tsquery.
     */
    self.getTsquery = function() {
      return self.getViewModel().tsquery();
    };

    self.getBoundTsquery = function() {
      return self.getViewModel().getBoundTsquery();
    };

    self.hasCharts = function() {
      return self.visualizer.hasCharts();
    };

    /**
     * Removes this plotContainer from the DOM.
     */
    self.remove = function() {
      self.unsubscribe();
      $container.remove();
    };

    self.unsubscribe = function() {
      self.visualizer.unsubscribe();
    };

    self.initialize();
  };

  return PlotContainer;
});
